﻿///////////////////////////////////////////////////////////////////////////////
// Activity: program.activity
// Diagram service implementation
///////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Xml;

using ccr = Microsoft.Ccr.Core;
using dss = Microsoft.Dss.Core;
using dssa = Microsoft.Dss.Core.Attributes;
using dssh = Microsoft.Dss.Core.DsspHttp;
using dssm = Microsoft.Dss.ServiceModel.DsspServiceBase;
using dssp = Microsoft.Dss.ServiceModel.Dssp;
using soap = W3C.Soap;

using submgr = Microsoft.Dss.Services.SubscriptionManager;
using drive = Microsoft.Robotics.Services.Drive.Proxy;
using xinputgamepad = Microsoft.Robotics.Services.Sample.XInputGamepad.Proxy;
using touchsensor = Microsoft.Robotics.Services.Sample.Lego.Nxt.TouchSensor.Proxy;

namespace Robotics.Application.Diagram
{
    [DisplayName("Application")]
    [Description("A user defined activity.")]
    [dssa.Contract(Contract.Identifier)]
    public class DiagramService : dssm.DsspServiceBase
    {
        // Service state
        [dssa.InitialStatePartner(Optional = true)]
        private DiagramState _state;

        // Service operations port
        [dssa.ServicePort("/Application", AllowMultipleInstances = false)]
        private DiagramOperations _mainPort = new DiagramOperations();

        #region Partner services

        [dssa.Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = dssa.PartnerCreationPolicy.CreateAlways)]
        private submgr.SubscriptionManagerPort _subMgr = new submgr.SubscriptionManagerPort();

        // Partner: XInputController, Contract: http://schemas.microsoft.com/robotics/2006/09/xinputgamepad.html
        [dssa.Partner("XInputController", Contract = xinputgamepad.Contract.Identifier, CreationPolicy = dssa.PartnerCreationPolicy.UsePartnerListEntry)]
        xinputgamepad.XInputGamepadOperations _xInputControllerPort = new xinputgamepad.XInputGamepadOperations();
        xinputgamepad.XInputGamepadOperations _xInputControllerNotify = new xinputgamepad.XInputGamepadOperations();

        // Partner: LegoNXTTouchSensorv2, Contract: http://schemas.microsoft.com/robotics/2007/07/lego/nxt/touchsensor.html
        [dssa.Partner("LegoNXTTouchSensorv2", Contract = touchsensor.Contract.Identifier, CreationPolicy = dssa.PartnerCreationPolicy.UsePartnerListEntry)]
        touchsensor.TouchSensorOperations _legoNXTTouchSensorv2Port = new touchsensor.TouchSensorOperations();
        touchsensor.TouchSensorOperations _legoNXTTouchSensorv2Notify = new touchsensor.TouchSensorOperations();

        // Partner: LegoNXTDrivev2, Contract: http://schemas.microsoft.com/robotics/2006/05/drive.html
        [dssa.Partner("LegoNXTDrivev2", Contract = drive.Contract.Identifier, CreationPolicy = dssa.PartnerCreationPolicy.UsePartnerListEntry)]
        drive.DriveOperations _legoNXTDrivev2Port = new drive.DriveOperations();

        #endregion

        public DiagramService(dssp.DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        protected override void Start()
        {
            // If there was no initial state partner, then the service state will be null.
            if (_state == null)
            {
                // The state MUST be created before the service starts processing messages.
                _state = new DiagramState();
            }

            // The rest of the start process requires the ability to wait for responses from
            // services and from the start handler (if any). So execution now proceeds in an
            // iterator function.
            SpawnIterator(DoStart);
        }

        private IEnumerator<ccr.ITask> DoStart()
        {
            soap.Fault fault = null;

            // Subscribe to partners

            yield return ccr.Arbiter.Choice(
                _xInputControllerPort.Subscribe(_xInputControllerNotify),
                EmptyHandler,
                delegate(soap.Fault f)
                {
                    fault = f;
                }
            );

            if (fault != null)
            {
                LogError("Failed to subscribe to program.XInputGamepadService.XInputController", fault);
                StartFailed();
                yield break;
            }

            yield return ccr.Arbiter.Choice(
                _legoNXTTouchSensorv2Port.Subscribe(_legoNXTTouchSensorv2Notify),
                EmptyHandler,
                delegate(soap.Fault f)
                {
                    fault = f;
                }
            );

            if (fault != null)
            {
                LogError("Failed to subscribe to program.NxtTouchSensor.LegoNXTTouchSensorv2", fault);
                StartFailed();
                yield break;
            }

            // Start operation handlers and insert into directory service.
            StartHandlers();

            // Add notifications to the main interleave
            base.MainPortInterleave.CombineWith(
                new ccr.Interleave(
                    new ccr.ExclusiveReceiverGroup(
                        ccr.Arbiter.ReceiveWithIterator<xinputgamepad.ButtonsChanged>(true, _xInputControllerNotify, XInputControllerButtonsChangedHandler),
                        ccr.Arbiter.ReceiveWithIterator<touchsensor.TouchSensorUpdate>(true, _legoNXTTouchSensorv2Notify, LegoNXTTouchSensorv2TouchSensorUpdateHandler)
                    ),
                    new ccr.ConcurrentReceiverGroup(
                        ccr.Arbiter.ReceiveWithIterator<xinputgamepad.ThumbsticksChanged>(true, _xInputControllerNotify, XInputControllerThumbsticksChangedHandler)
                    )
                )
            );

            yield break;
        }

        private void StartHandlers()
        {
            // Activate message handlers for this service and insert into the directory.
            base.Start();
        }

        #region Standard DSS message handlers

        [dssa.ServiceHandler(dssa.ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ccr.ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        [dssa.ServiceHandler(dssa.ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ccr.ITask> HttpGetHandler(dssh.HttpGet httpGet)
        {
            httpGet.ResponsePort.Post(new dssh.HttpResponseType(_state));
            yield break;
        }

        [dssa.ServiceHandler(dssa.ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ccr.ITask> ReplaceHandler(Replace replace)
        {
            _state = replace.Body;

            replace.ResponsePort.Post(dssp.DefaultReplaceResponseType.Instance);
            base.SendNotification<Replace>(_subMgr, replace);

            yield break;
        }

        [dssa.ServiceHandler(dssa.ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ccr.ITask> SubscribeHandler(Subscribe subscribe)
        {
            dssp.SubscribeRequestType request = subscribe.Body;

            yield return ccr.Arbiter.Choice(
                SubscribeHelper(_subMgr, request, subscribe.ResponsePort),
                delegate(ccr.SuccessResult success)
                {
                    base.SendNotification<Replace>(_subMgr, request.Subscriber, _state);
                },
                delegate(Exception e) { }
            );
        }

        #endregion

        #region Custom message handlers

        [dssa.ServiceHandler(dssa.ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ccr.ITask> ActionHandler(Action message)
        {
            // Empty handler. Respond with default response type.
            message.ResponsePort.Post(new ActionResponse());
            yield break;
        }

        #endregion

        #region Notification message handlers

        IEnumerator<ccr.ITask> XInputControllerButtonsChangedHandler(xinputgamepad.ButtonsChanged message)
        {
            OnXInputControllerButtonsChangedHandler handler = new OnXInputControllerButtonsChangedHandler(this, Environment.TaskQueue);
            SpawnIterator(message, handler.RunHandler);
            yield return ccr.Arbiter.Receive(false, handler.Complete, EmptyHandler);

            if (handler.StateChanged == true)
            {
                base.SendNotification<Replace>(_subMgr, _state);
            }
        }

        IEnumerator<ccr.ITask> XInputControllerThumbsticksChangedHandler(xinputgamepad.ThumbsticksChanged message)
        {
            OnXInputControllerThumbsticksChangedHandler handler = new OnXInputControllerThumbsticksChangedHandler(this, Environment.TaskQueue);
            return handler.RunHandler(message);
        }

        IEnumerator<ccr.ITask> LegoNXTTouchSensorv2TouchSensorUpdateHandler(touchsensor.TouchSensorUpdate message)
        {
            OnLegoNXTTouchSensorv2TouchSensorUpdateHandler handler = new OnLegoNXTTouchSensorv2TouchSensorUpdateHandler(this, Environment.TaskQueue);
            SpawnIterator(message, handler.RunHandler);
            yield return ccr.Arbiter.Receive(false, handler.Complete, EmptyHandler);

            if (handler.StateChanged == true)
            {
                base.SendNotification<Replace>(_subMgr, _state);
            }
        }

        #endregion

        #region OnXInputControllerButtonsChangedHandler class

        class OnXInputControllerButtonsChangedHandler : HandlerBase
        {
            ///////////////////////////////////////////////////////////////////
            // if if{expr - set}else{}
            ///////////////////////////////////////////////////////////////////

            public OnXInputControllerButtonsChangedHandler(DiagramService service, ccr.DispatcherQueue queue)
                : base(service, queue)
            {
            }

            ///////////////////////////////////////////////////////////////////
            // if if{expr - set}else{}
            ///////////////////////////////////////////////////////////////////

            public IEnumerator<ccr.ITask> RunHandler(xinputgamepad.ButtonsChanged message)
            {

                if (message.Body.A)
                {
                    base.StateChanged = true;
                    State.MotorEnabled = true;
                }

                base.Complete.Post(ccr.EmptyValue.SharedInstance);
                yield break;
            }
        }

        #endregion

        #region OnXInputControllerThumbsticksChangedHandler class

        class OnXInputControllerThumbsticksChangedHandler : HandlerBase
        {
            ///////////////////////////////////////////////////////////////////
            // __use__.subscribe0
            // get - if0 if{&join}else{}
            // join
            // __use__0.snippet0.snippet.noop - [(__use__0.snippet0.snippet.expr0 - __use__0.snippet0.snippet.join)(__use__0.snippet0.snippet.expr - &__use__0.snippet0.snippet.join)] - __use__0.snippet0.call - __use__0.snippet0.call.iftype
            ///////////////////////////////////////////////////////////////////

            public OnXInputControllerThumbsticksChangedHandler(DiagramService service, ccr.DispatcherQueue queue)
                : base(service, queue)
            {

                // Register joins with base class to manage active count for incomplete joins.
                base.RegisterJoin(_joinAlphaPorts);
                // Activate Join handlers
                Activate(ccr.Arbiter.MultiplePortReceive(true, _joinAlphaPorts, _joinAlphaHandler));
            }

            ///////////////////////////////////////////////////////////////////
            // __use__.subscribe0
            // join
            // get - if0 if{&join}else{}
            ///////////////////////////////////////////////////////////////////

            public IEnumerator<ccr.ITask> RunHandler(xinputgamepad.ThumbsticksChanged message)
            {
                Increment();


                Increment();
                _joinAlphaPorts[1].Post(message.Body);


                DiagramState a = new DiagramState();
                a.MotorEnabled = State.MotorEnabled;

                if (a.MotorEnabled)
                {
                    Increment();
                    _joinAlphaPorts[0].Post(a);
                }

                Decrement();

                yield return WaitUntilComplete();
            }

            ///////////////////////////////////////////////////////////////////
            // __use__0.snippet0.snippet.noop - [(__use__0.snippet0.snippet.expr0 - __use__0.snippet0.snippet.join)(__use__0.snippet0.snippet.expr - &__use__0.snippet0.snippet.join)] - __use__0.snippet0.call - __use__0.snippet0.call.iftype
            ///////////////////////////////////////////////////////////////////

            void _joinAlphaHandler(object[] args)
            {
                JoinAlpha message = new JoinAlpha(args);
                drive.SetDrivePowerRequest request = new drive.SetDrivePowerRequest();
                request.RightWheelPower = (double)message.msg0.RightY;
                request.LeftWheelPower = (double)message.msg0.LeftY;
                LegoNXTDrivev2Port.SetDrivePower(request);

                Decrement(args.Length);
            }

            class JoinAlpha
            {
                public DiagramState msg;
                public xinputgamepad.Thumbsticks msg0;

                public JoinAlpha()
                {
                }

                public JoinAlpha(object[] args)
                {
                    msg = (DiagramState)args[0];
                    msg0 = (xinputgamepad.Thumbsticks)args[1];
                }
            }

            ccr.Port<object>[] _joinAlphaPorts = new ccr.Port<object>[2]{
                new ccr.Port<object>(),
                new ccr.Port<object>()
            };

            class JoinBeta
            {
                public string LeftWheelPower;
                public string RightWheelPower;

                public JoinBeta()
                {
                }

                public JoinBeta(object[] args)
                {
                    LeftWheelPower = args[0].ToString();
                    RightWheelPower = args[1].ToString();
                }
            }

            ccr.Port<object>[] _joinBetaPorts = new ccr.Port<object>[2]{
                new ccr.Port<object>(),
                new ccr.Port<object>()
            };
        }

        #endregion

        #region OnLegoNXTTouchSensorv2TouchSensorUpdateHandler class

        class OnLegoNXTTouchSensorv2TouchSensorUpdateHandler : HandlerBase
        {
            ///////////////////////////////////////////////////////////////////
            // if1 if{(expr0 - set0) (__use__2.snippet.snippet.noop - [(__use__2.snippet.snippet.expr0 - __use__2.snippet.snippet.join)(__use__2.snippet.snippet.expr - &__use__2.snippet.snippet.join)] - __use__2.snippet.call - __use__2.snippet.call.iftype)}else{}
            ///////////////////////////////////////////////////////////////////

            public OnLegoNXTTouchSensorv2TouchSensorUpdateHandler(DiagramService service, ccr.DispatcherQueue queue)
                : base(service, queue)
            {
            }

            ///////////////////////////////////////////////////////////////////
            // if1 if{(expr0 - set0) (__use__2.snippet.snippet.noop - [(__use__2.snippet.snippet.expr0 - __use__2.snippet.snippet.join)(__use__2.snippet.snippet.expr - &__use__2.snippet.snippet.join)] - __use__2.snippet.call - __use__2.snippet.call.iftype)}else{}
            ///////////////////////////////////////////////////////////////////

            public IEnumerator<ccr.ITask> RunHandler(touchsensor.TouchSensorUpdate message)
            {

                if (message.Body.TouchSensorOn)
                {
                    base.StateChanged = true;
                    State.MotorEnabled = false;

                    drive.SetDrivePowerRequest request = new drive.SetDrivePowerRequest();
                    LegoNXTDrivev2Port.SetDrivePower(request);
                }

                base.Complete.Post(ccr.EmptyValue.SharedInstance);
                yield break;
            }

            class JoinAlpha
            {
                public string LeftWheelPower;
                public string RightWheelPower;

                public JoinAlpha()
                {
                }

                public JoinAlpha(object[] args)
                {
                    LeftWheelPower = args[0].ToString();
                    RightWheelPower = args[1].ToString();
                }
            }

            ccr.Port<object>[] _joinAlphaPorts = new ccr.Port<object>[2]{
                new ccr.Port<object>(),
                new ccr.Port<object>()
            };
        }

        #endregion

        #region Handler utility base class

        class HandlerBase : ccr.CcrServiceBase
        {
            ccr.Port<ccr.EmptyValue> _complete = new ccr.Port<ccr.EmptyValue>();
            ccr.Port<ccr.EmptyValue> _shutdown = new ccr.Port<ccr.EmptyValue>();
            List<ccr.Port<object>[]> _joins = new List<ccr.Port<object>[]>();
            DiagramService _service;
            int _count;

            public HandlerBase(DiagramService service, ccr.DispatcherQueue queue)
                : base(queue)
            {
                _service = service;
            }

            protected DiagramState State
            {
                get { return _service._state; }
            }

            bool _stateChanged;

            public bool StateChanged
            {
                get { return _stateChanged; }
                set { _stateChanged = value; }
            }

            protected xinputgamepad.XInputGamepadOperations XInputControllerPort
            {
                get { return _service._xInputControllerPort; }
            }

            protected touchsensor.TouchSensorOperations LegoNXTTouchSensorv2Port
            {
                get { return _service._legoNXTTouchSensorv2Port; }
            }

            protected drive.DriveOperations LegoNXTDrivev2Port
            {
                get { return _service._legoNXTDrivev2Port; }
            }

            public ccr.Port<ccr.EmptyValue> Complete
            {
                get { return _complete; }
            }

            protected void Increment()
            {
                System.Threading.Interlocked.Increment(ref _count);
            }

            protected void Decrement()
            {
                int offset = 0;

                foreach (ccr.Port<object>[] join in _joins)
                {
                    bool complete = true;
                    int joinCount = 0;

                    foreach (ccr.Port<object> sink in join)
                    {
                        if (sink.ItemCount != 0)
                        {
                            joinCount += sink.ItemCount;
                        }
                        else
                        {
                            complete = false;
                        }
                    }
                    if (complete == false)
                    {
                        offset += joinCount;
                    }
                }

                if (System.Threading.Interlocked.Decrement(ref _count) <= offset)
                {
                    _shutdown.Post(ccr.EmptyValue.SharedInstance);
                    _complete.Post(ccr.EmptyValue.SharedInstance);
                }
            }

            protected void Decrement(int count)
            {
                for(int i = 0; i < count; i++)
                {
                    Decrement();
                }
            }

            protected void RegisterJoin(ccr.Port<object>[] join)
            {
                _joins.Add(join);
            }

            protected ccr.ITask WaitUntilComplete()
            {
                return ccr.Arbiter.Receive(false, _shutdown, NullDelegate);
            }

            protected static void NullDelegate(ccr.EmptyValue token)
            {
            }

            protected void SendNotification<T>(object notification)
                where T : dssp.DsspOperation, new()
            {
                _service.SendNotification<T>(_service._subMgr, notification);
            }

            protected void FaultHandler(soap.Fault fault, string msg)
            {
                _service.LogError(null, msg, fault);
            }

            protected static string Stringize(object obj)
            {
                if (obj == null)
                {
                    return string.Empty;
                }
                else
                {
                    return obj.ToString();
                }
            }

            protected void UnhandledResponse(W3C.Soap.Fault fault)
            {
                _service.LogError("Unhandled fault response from partner service", fault);
                Decrement();
            }

            protected void UnhandledResponse<T>(T response)
            {
                Decrement();
            }

            #region Type Cast Functions

            protected static bool BoolCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToBoolean(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return false;
            }

            protected static byte ByteCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToByte(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static char CharCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToChar(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return '\0';
            }

            protected static decimal DecimalCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToDecimal(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static double DoubleCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToDouble(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static float FloatCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToSingle(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static int IntCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToInt32(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static long LongCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToInt64(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static sbyte SByteCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToSByte(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static short ShortCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToInt16(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static uint UIntCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToUInt32(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static ulong ULongCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToUInt64(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            protected static ushort UShortCast(object obj)
            {
                if (obj != null && obj is IConvertible)
                {
                    try
                    {
                        return ((IConvertible)obj).ToUInt16(System.Globalization.CultureInfo.InvariantCulture);
                    }
                    catch (FormatException)
                    {
                    }
                }
                return 0;
            }

            #endregion

            #region List Functions

            protected List<T> ListAdd<T>(List<T> list, T item)
            {
                List<T> output = new List<T>(list);
                output.Add(item);
                return output;
            }

            protected List<T> ListConcat<T>(List<T> head, List<T> tail)
            {
                List<T> output = new List<T>(head);
                output.AddRange(tail);
                return output;
            }

            protected int ListIndex<T>(List<T> list, T item)
            {
                return list.IndexOf(item);
            }

            protected List<T> ListRemove<T>(List<T> list, int index)
            {
                List<T> output = new List<T>(list);
                output.RemoveAt(index);
                return output;
            }

            protected List<T> ListReverse<T>(List<T> list)
            {
                List<T> output = new List<T>(list);
                output.Reverse();
                return output;
            }

            protected List<T> ListSort<T>(List<T> list)
            {
                List<T> output = new List<T>(list);
                output.Sort();
                return output;
            }

            protected List<T> ListInsert<T>(List<T> list, T item, int index)
            {
                List<T> output = new List<T>(list);
                output.Insert(index, item);
                return output;
            }

            #endregion
        }

        #endregion
    }
}
